from ._bindings import AssembleResult, Cache, Component, Enhancements

AssembleResult.__module__ = __name__
Cache.__module__ = __name__
Component.__module__ = __name__
Enhancements.__module__ = __name__
